package captcha

import (
	"context"
	"strings"
	"time"

	"github.com/mojocn/base64Captcha"
	"github.com/redis/go-redis/v9"
	"gitlab.wanzhuangkj.com/tush/xpkg/xerrors/xerror"
)

const (
	NotFoundPlaceholder = "*"
)

type Captcha struct {
	captcha *base64Captcha.Captcha
}

func (r *Captcha) Verify(id, answer string, clear bool) bool {
	return r.captcha.Store.Verify(id, answer, clear)
}

func (r *Captcha) Generate(ctx context.Context) (id, b64s, answer string, err error) {
	return r.captcha.Generate()
}

type RedisStore struct {
	client     *redis.Client // 支持单节点和集群模式
	prefix     string        // Key前缀隔离
	expiration time.Duration // 验证码有效期
}

func NewRedisStore(redisClient *redis.Client, prefix string) *RedisStore {
	return &RedisStore{
		client:     redisClient,
		prefix:     prefix,
		expiration: 10 * time.Minute,
	}
}

func (rs *RedisStore) Set(id string, value string) error {
	ctx := context.Background()
	fullValue := value + "|" + time.Now().Format(time.RFC3339)
	return rs.client.Set(ctx, rs.prefix+id, fullValue, rs.expiration).Err()
}

func (rs *RedisStore) Get(id string, clear bool) string {
	ctx := context.Background()
	key := rs.prefix + id
	val, err := rs.client.Get(ctx, key).Result()
	if err != nil {
		if xerror.Is(err, redis.Nil) {
			return ""
		}
		panic(xerror.New(err.Error()))
	}
	if clear {
		go rs.client.Del(ctx, key)
	}
	parts := strings.Split(val, "|")
	if len(parts) != 2 {
		return ""
	}
	if parts[0] == NotFoundPlaceholder { // 这块逻辑应该用不到
		return ""
	}
	return parts[0]
}

func (r *RedisStore) Verify(id, answer string, clear bool) bool {
	stored := r.Get(id, clear)
	return strings.EqualFold(stored, answer)
}

func NewCaptcha(redisClient *redis.Client) *Captcha {
	store := NewRedisStore(redisClient, "captcha:")
	driver := &base64Captcha.DriverDigit{
		Height:   80,
		Width:    240,
		Length:   5,
		MaxSkew:  0.7,
		DotCount: 80,
	}
	return &Captcha{
		captcha: base64Captcha.NewCaptcha(driver, store),
	}
}
